import unittest

from satella.coding.optionals import call_if_nnone, iterate_if_nnone, Optional, extract_optional


class TestOptionals(unittest.TestCase):

    def test_extract_optional(self):
        b = Optional(None)
        self.assertIs(extract_optional(b.test.value), None)
        self.assertEqual(extract_optional(5), 5)

    def test_optional(self):
        b = Optional(None)
        self.assertFalse(b.consumer.cancel().result())

    def test_iterate(self):
        b = list(iterate_if_nnone([1, 2, 3]))
        c = list(iterate_if_nnone(None))
        self.assertFalse(c)
        self.assertTrue(b)

    def test_optional_lambda(self):
        Optional(None)()
        Optional(lambda: 5)()

    def test_object(self):
        class Obj:
            def __init__(self):
                self.a = 5

        obj = Obj()
        opt = Optional(obj)
        Optional(None).a
        self.assertEqual(opt.a, 5)
        opt.a = 6
        self.assertEqual(obj.a, 6)
        self.assertEqual(opt.a, 6)
        del opt.a
        self.assertRaises(AttributeError, lambda: obj.a)
        self.assertRaises(AttributeError, lambda: opt.a)

    def test_list(self):
        a = [1, 2, 3]
        opt = Optional(a)
        self.assertEqual(opt[0], 1)
        del opt[0]
        self.assertEqual(len(opt), 2)
        opt[0] = 4
        self.assertEqual(opt[0], 4)

    def test_optional_eq(self):
        class Opt:
            a = 5
            b = [2]

        a = Opt()
        self.assertEqual(Optional(a).a, 5)
        self.assertIn(2, Optional(a).b)
        self.assertNotEqual(Optional(None).a, 5)
        self.assertNotIn(2, Optional(None).b)
        self.assertEqual(Optional(5), Optional(5))

    def test_optional(self):
        self.assertIsNone(call_if_nnone(None))
        b = call_if_nnone(lambda y: y, 5)
        self.assertIsNotNone(b)
        self.assertEqual(5, b)



from satella.coding.concurrent import CallableGroup

from satella.cassandra import parallel_for, wrap_future
import unittest


class TestCassandra(unittest.TestCase):
    def test_wrap_future(self):
        class MockCassandraFuture:
            def __init__(self):
                self.value = None
                self.callbacks = CallableGroup()
                self.errbacks = CallableGroup()

            def add_callback(self, callback):
                self.callbacks.add(callback)

            def add_errback(self, errback):
                self.errbacks.add(errback)

            def set_result(self, x):
                self.value = x
                if isinstance(x, Exception):
                    self.errbacks(x)
                else:
                    self.callbacks(x)

        mcf = MockCassandraFuture()
        wrapped = wrap_future(mcf)
        self.assertEqual(id(wrapped), id(wrap_future(wrapped)))
        a = {}

        def on_done(fut):
            if fut.exception() is None:
                a['success'] = True
            else:
                a['failure'] = True

        wrapped.add_done_callback(on_done)

        mcf.set_result(None)
        self.assertTrue(a['success'])

        mcf = MockCassandraFuture()
        wrapped = wrap_future(mcf)
        wrapped.add_done_callback(on_done)
        mcf.set_result(Exception())
        self.assertTrue(a['failure'])

    def test_parallel_for(self):
        class Cursor:
            def __init__(self):
                self.execute_times_called = 0
                self.result_times_called = 0
                self.execute_without_args_called = 0

            def result(self):
                self.result_times_called += 1
                return []

            def execute_async(self, query, args=None):
                if args is None:
                    self.execute_without_args_called += 1
                self.execute_times_called += 1
                return self

        cur = Cursor()
        list(parallel_for(cur, 'SELECT * FROM table', [(1,), (2, ), (3, )]))

        self.assertEqual(cur.execute_times_called, 3)
        self.assertEqual(cur.result_times_called, 3)

        list(parallel_for(cur, ['SELECT * FROM table',
                                'SELECT * FROM table2',
                                'SELECT * FROM table3'], [(1,), (2, ), (3, )]))

        self.assertEqual(cur.execute_times_called, 6)
        self.assertEqual(cur.result_times_called, 6)

        list(parallel_for(cur, ['SELECT * FROM table',
                                'SELECT * FROM table2',
                                'SELECT * FROM table3'], [None, None, None]))

        self.assertEqual(cur.execute_without_args_called, 3)




import unittest


from satella.exceptions import BaseSatellaError, CustomException, CodedCustomException, \
    ImpossibleError, CodedCustomException


class TestExceptions(unittest.TestCase):
    def test_advanced_class_hierarchy(self):
        class A(CodedCustomException):
            code = None  # type: int
            message = ''  # type: str

        class B(A):
            code = 2
            message = 'not readed error'

        b = B()

    def test_coded_custom_exception_no_args(self):
        CodedCustomException()

    def test_impossible_error(self):
        try:
            raise ImpossibleError()
        except Exception:
            self.fail('failed')
        except BaseException:
            pass
        else:
            self.fail('failed')

    def test_exception_kwargs(self):
        e = BaseSatellaError('hello world', label='value')
        self.assertIn("label='value'", repr(e))

    def test_except_inherited(self):
        class InheritedError(CustomException):
            pass

        try:
            raise InheritedError('message', 'arg1', 'arg2')
        except CustomException as e:
            self.assertEqual(str(e), "InheritedError('message', 'arg1', 'arg2')")
        else:
            self.fail()

    def test_coded_exception(self):
        class BaseError(CodedCustomException):
            pass

        class Base2Error(BaseError):
            code = 2

        self.assertIsInstance(BaseError('message', 2), CodedCustomException)
        self.assertIsInstance(Base2Error('message'), CodedCustomException)
        self.assertNotIsInstance(CodedCustomException('message', 2), Base2Error)
        self.assertIsInstance(BaseError('message', 2), Base2Error)
        self.assertIsInstance(Base2Error('message'), Base2Error)

        class DifferentHierarchy(CodedCustomException):
            pass

        self.assertNotIsInstance(DifferentHierarchy('message', 2), Base2Error)



import threading
import unittest

from satella.coding.environment import Context


class TestEnvs(unittest.TestCase):

    def test_envs(self):
        ctxt = Context.get()
        ctxt.value = 5
        self.assertEqual(ctxt.value, 5)
        with Context() as new_ctxt:
            self.assertEqual(new_ctxt.value, 5)

    def test_threads(self):
        ctxt = Context.get()
        ctxt.value = 2
        with ctxt:
            def thread():
                with Context() as new_ctxt:
                    self.assertEqual(new_ctxt.value, 2)
                    new_ctxt.value2 = 2
            thr = threading.Thread(target=thread)
            thr.start()
            thr.join()
            self.assertRaises(AttributeError, lambda: ctxt.value2)

    def test_push(self):
        ctxt = Context.get()
        ctxt.value = 5
        self.assertEqual(ctxt.value, 5)
        with Context() as new_ctxt:
            self.assertEqual(new_ctxt.value, 5)
            with Context() as new_ctxt2:
                new_ctxt2.push_up('value', 6)
            self.assertEqual(new_ctxt.value, 6)

        self.assertEqual(ctxt.value, 5)

    def test_delete_envs(self):
        ctxt = Context.get()
        ctxt.value = 5
        self.assertEqual(ctxt.value, 5)
        with Context() as new_ctxt:
            del new_ctxt.value
            self.assertRaises(AttributeError, lambda: new_ctxt.value)
        self.assertEqual(ctxt.value, 5)

    def test_nesting(self):
        with Context() as ctxt:
            ctxt.value = 55
            with Context() as new_ctxt:
                new_ctxt.value = 66
                assert new_ctxt.value == 66
            assert ctxt.value == 55



import unittest
import sys
from satella.processes import call_and_return_stdout


class TestProcesses(unittest.TestCase):
    @unittest.skipIf('win' in sys.platform, 'Running on Windows')
    def test_timeouts(self):
        self.assertRaises(TimeoutError, lambda: call_and_return_stdout('sleep 5',
                                                                       shell=True, timeout=3))

    @unittest.skipIf('win' in sys.platform, 'Running on Windows')
    def test_return_stdout(self):
        output = call_and_return_stdout('ls .', shell=True, timeout=3, expected_return_code=0)
        self.assertIn(b'pyproject.toml', output)

    @unittest.skipIf('win' in sys.platform or sys.version_info.minor < 6, 'Running on Windows or Python 3.5')
    def test_return_encoding(self):
        output = call_and_return_stdout('ls .', shell=True, encoding='utf8', timeout=3)
        self.assertIn('pyproject.toml', output)



import io
import os
from os.path import join
import tempfile
import unittest
import shutil
from satella.files import read_re_sub_and_write, find_files, split, read_in_file, write_to_file, \
    write_out_file_if_different, make_noncolliding_name, try_unlink, DevNullFilelikeObject, \
    read_lines, AutoflushFile


def putfile(path: str) -> None:
    with open(path, 'wb') as f_out:
        f_out.write(b'\x32')


class TestFiles(unittest.TestCase):

    def test_read_nonexistent_file(self):
        self.assertRaises(FileNotFoundError, lambda: read_in_file('moot'))

    def test_autoflush_file(self):
        af = AutoflushFile('test3.txt', 'w+', encoding='utf-8')
        try:
            af.write('test')
            assert read_in_file('test3.txt', encoding='utf-8') == 'test'
            af.write('test2')
            assert read_in_file('test3.txt', encoding='utf-8') == 'testtest2'
            af.truncate(4)
            assert read_in_file('test3.txt', encoding='utf-8') == 'test'
        finally:
            af.close()
            try_unlink('test3.txt')

        af = AutoflushFile('test3.txt', 'w', encoding='utf-8')
        os.unlink('test3.txt')
        af.close()
        self.assertRaises(ValueError, lambda: af.write('test'))

    def test_read_lines(self):
        lines = read_lines('LICENSE')
        self.assertTrue(all(lines))

        with open('test.tmp', 'w') as f_out:
            f_out.write('''line1
            line2
            line3
            
            ''')

        self.assertEqual(read_lines('test.tmp'), ['line1', 'line2', 'line3'])

    def test_devnullfilelikeobject(self):
        null = DevNullFilelikeObject()
        self.assertEqual(null.write('ala'), 3)
        assert null.seek(0) == 0
        assert null.tell() == 0
        assert null.seekable()
        assert null.truncate(0) == 0
        self.assertRaises(TypeError, lambda: null.write(b'ala'))
        self.assertEqual(null.read(), '')
        self.assertEqual(null.read(7), '')
        null.flush()
        null.close()
        self.assertRaises(ValueError, lambda: null.write('test'))
        self.assertRaises(ValueError, lambda: null.flush())
        self.assertRaises(ValueError, lambda: null.read())
        null.close()

    def test_devnullfilelikeobject_2(self):
        null = DevNullFilelikeObject(binary=True)
        null.write(b'test')
        self.assertRaises(TypeError, lambda: null.write('ala'))

        null = DevNullFilelikeObject()
        self.assertRaises(TypeError, lambda: null.write(b'test'))
        null.write('test')

    def test_devnullfilelikeobject_3(self):
        null = DevNullFilelikeObject(binary=True)
        self.assertEqual(null.write(b'ala'), 3)
        assert null.seek(0) == 0
        assert null.tell() == 0
        assert null.seekable()
        assert null.truncate(0) == 0
        self.assertEqual(null.write(b'ala'), 3)
        self.assertEqual(null.read(), b'')
        self.assertEqual(null.read(7), b'')
        null.flush()
        null.close()
        self.assertRaises(ValueError, lambda: null.write('test'))
        self.assertRaises(ValueError, lambda: null.flush())
        self.assertRaises(ValueError, lambda: null.read())
        null.close()

    def try_directory(self):
        os.system('mkdir test')
        self.assertRaises(FileNotFoundError, lambda: read_in_file('test'))
        self.assertEqual(b'test', read_in_file('test', default=b'test'))
        os.system('rm -rf test')
        self.assertRaises(FileNotFoundError, lambda: read_in_file('test'))
        self.assertEqual(b'test', read_in_file('test', default=b'test'))

    def test_make_noncolliding_name(self):
        with open('test.txt', 'w') as f_out:
            f_out.write('test')
        self.assertEqual(make_noncolliding_name('test.txt'), 'test.1.txt')
        self.assertTrue(try_unlink('test.txt'))
        self.assertFalse(try_unlink('test.txt'))

    def test_write_out_file_if_different(self):
        try:
            os.unlink('test')
        except FileNotFoundError:
            pass
        try:
            self.assertTrue(write_out_file_if_different('test', 'test', 'UTF-8'))
            self.assertFalse(write_out_file_if_different('test', 'test', 'UTF-8'))
        finally:
            os.unlink('test')

    def test_read_in_and_write_to(self):
        data = 'żażółć gęślą jaźń'
        write_to_file('temp.tmp', data, 'UTF-8')
        data2 = read_in_file('temp.tmp', 'UTF-8')
        self.assertEqual(data, data2)

    def test_split(self):
        self.assertIn(split('c:/windows/system32/system32.exe'), [['c:', 'windows', 'system32',
                                                                     'system32.exe'],
                                                                  ['c:/', 'windows', 'system32',
                                                                   'system32.exe']
                                                                  ])
        self.assertEqual(split('~/user/something/./else'), ['~', 'user', 'something', '.',
                                                            'else'])

    def setUp(self) -> None:
        self.filename = tempfile.mktemp()

    def tearDown(self) -> None:
        try:
            os.unlink(self.filename)
        except OSError:
            pass

    def test_find_files(self):
        directory = tempfile.mkdtemp()
        os.mkdir(join(directory, 'test'))
        putfile(join(directory, 'test', 'test.txt'))
        os.mkdir(join(directory, 'test', 'test'))
        putfile(join(directory, 'test', 'test', 'test.txt'))
        putfile(join(directory, 'test', 'test', 'test2.txt'))
        self.assertEqual(set(find_files(directory, r'(.*)test(.*)\.txt',
                                         apply_wildcard_to_entire_path=True)), {
            join(directory, 'test', 'test.txt'),
            join(directory, 'test', 'test', 'test.txt'),
            join(directory, 'test', 'test', 'test2.txt')
        })

        self.assertEqual(set(find_files(directory, r'(.*)\.txt')), {
            join(directory, 'test', 'test.txt'),
            join(directory, 'test', 'test', 'test.txt'),
            join(directory, 'test', 'test', 'test2.txt')
        })
        shutil.rmtree(directory)

    def test_read_re_sub_and_write(self):
        with open(self.filename, 'w') as f_out:
            f_out.write('RE')

        read_re_sub_and_write(self.filename, r'RE', 'sub')

        with open(self.filename, 'r') as f_in:
            self.assertEqual('sub', f_in.read())

    def test_read_re_sub_and_write_2(self):
        with open(self.filename, 'w') as f_out:
            f_out.write('{2} {3}')

        read_re_sub_and_write(self.filename, r'{(.*?)}', lambda match_obj: match_obj.group(1))

        with open(self.filename, 'r') as f_in:
            self.assertEqual('2 3', f_in.read())



import collections
import threading
import typing as tp

from satella.coding.concurrent.thread import Condition
from satella.coding.recast_exceptions import rethrow_as
from satella.coding.typing import T
from satella.exceptions import WouldWaitMore, Empty
from satella.time.measure import measure


class PeekableQueue(tp.Generic[T]):
    """
    A thread-safe FIFO queue that supports peek()ing for elements.
    """
    __slots__ = ('queue', 'lock', 'inserted_condition')

    def __init__(self):
        super().__init__()
        self.queue = collections.deque()
        self.lock = threading.Lock()
        self.inserted_condition = Condition()

    def put(self, item: T) -> None:
        """
        Add an element to the queue

        :param item: element to add
        """
        with self.lock:
            self.queue.append(item)
        self.inserted_condition.notify()

    def put_many(self, items: tp.Sequence[T]) -> None:
        """
        Put multiple items

        :param items: sequence of items to put
        """
        with self.lock:
            items_count = 0
            for item in items:
                items_count += 1
                self.queue.append(item)
        self.inserted_condition.notify(items_count)

    def __get_timeout_none(self, item_getter):
        while True:
            self.lock.release()
            self.inserted_condition.wait()
            self.lock.acquire()
            if len(self.queue):
                try:
                    return item_getter(self.queue)
                finally:
                    self.lock.release()

    def __get_timeout(self, item_getter, timeout):
        with measure(timeout=timeout) as measurement:
            while not measurement.timeouted:
                self.lock.release()
                # raises WouldWaitMore
                self.inserted_condition.wait(timeout=measurement.time_remaining)
                self.lock.acquire()
                if len(self.queue):
                    try:
                        return item_getter(self.queue)
                    finally:
                        self.lock.release()
            else:
                self.lock.release()
                raise Empty('queue is empty')

    @rethrow_as(WouldWaitMore, Empty)
    def __get(self, timeout, item_getter) -> T:
        self.lock.acquire()
        if len(self.queue):
            # Fast path
            try:
                return item_getter(self.queue)
            finally:
                self.lock.release()
        else:
            if timeout is None:
                return self.__get_timeout_none(item_getter)
            else:
                return self.__get_timeout(item_getter, timeout)

    def get(self, timeout: tp.Optional[float] = None) -> T:
        """
        Get an element.

        :param timeout: maximum amount of seconds to wait. Default value of None
            means wait as long as necessary
        :return: the item
        :raise Empty: queue was empty
        """
        return self.__get(timeout, lambda queue: queue.popleft())

    def peek(self, timeout: tp.Optional[float] = None) -> T:
        """
        Get an element without removing it from the top of the queue.

        :param timeout: maximum amount of seconds to wait. Default value of None
            means wait as long as necessary
        :return: the item
        :raise WouldWaitMore: timeout has expired
        """
        return self.__get(timeout, lambda queue: queue[0])

    def qsize(self) -> int:
        """
        Return the approximate size of the queue. Note, qsize() > 0 doesn’t
        guarantee that a subsequent get() will not block.
        :return: approximate size of the queue
        """
        return len(self.queue)



import warnings

__all__ = ['ExperimentalWarning', 'warnings', 'mark_temporarily_disabled']


class ExperimentalWarning(FutureWarning):
    """
    This feature is experimental!
    """


def mark_temporarily_disabled(reason: str = ''):
    """
    A decorator to mark a function unusable due to some forthcoming changes.

    A call to the function will raise NotImplementedError

    and a mention of this function in code will mark a function disabled for whatever reason.

    Usage:

    >>> @mark_temporarily_disabled('Due to analysis schema changes')
    >>> def serialize_report(analysis):
    >>>     ...
    """

    def outer(fun):
        def inner(*args, **kwargs):
            """
            Function temporarily disabled. Don't call it.
            """
            raise NotImplementedError(f'Function disabled due to {reason}')

        warnings.warn(f'Function {fun.__name__} temporarily disabled due to {reason}')
        return inner

    return outer



import unittest

from satella.coding.concurrent import call_in_separate_thread
from satella.opentracing import trace_exception, trace_future, trace_function, \
    set_maximum_traceback_length
from unittest import mock


class TestOpentracing(unittest.TestCase):
    def test_trace_function(self):
        slf = self

        class MockTracer:
            def __init__(self):
                self.start_active_span_called = True

            def start_active_span(self, *args, tags=None, **kwargs):
                slf.assertEqual(tags, {'a': 'b', 'c': 'd', 'e': 'f'})

                class MockScope(mock.Mock):
                    def __enter__(self):
                        return self

                    def __exit__(self, exc_type, exc_val, exc_tb):
                        return False
                self.start_active_span_called = True
                return MockScope()

        tracer = MockTracer()

        @trace_function(tracer, 'trace_me',
                        tags_factory= {'a': lambda: 'b',
                                       'c': lambda: 'd',
                                       'e': (lambda args: 'f')})
        def trace_me():
            pass

        trace_me()
        self.assertTrue(tracer.start_active_span_called)

    def test_trace_future_success(self):
        @call_in_separate_thread()
        def test_me():
            pass

        fut = test_me()
        span = mock.MagicMock()
        trace_future(fut, span)
        fut.result()
        fut.thread.join()
        self.assertTrue(span.finish.called)

    def test_trace_future_exception(self):
        @call_in_separate_thread()
        def fail_me():
            raise ValueError()

        fut = fail_me()
        span = mock.Mock()
        trace_future(fut, span)
        self.assertRaises(ValueError, fut.result)
        fut.thread.join()
        self.assertTrue(span.finish.called)

    def test_trace_exception_none(self):
        trace_exception(None)
        set_maximum_traceback_length(32768)

        span = mock.MagicMock()
        try:
            raise ValueError()
        except ValueError:
            trace_exception(span)

        self.assertTrue(span.set_tag.called)
        self.assertTrue(span.log_kv.called)



import typing as tp
import warnings

from satella.coding.structures import Proxy
from satella.coding.typing import V, T

__all__ = ['call_if_nnone', 'iterate_if_nnone', 'Optional', 'extract_optional']


def iterate_if_nnone(iterable: tp.Optional[tp.Iterable]) -> tp.Iterable:
    """
    Return a generator iterating over every element of iterable if it's not None.

    If it's None, return an empty generator

    :param iterable: iterable to iterate over
    :return: an empty generator if iterable is none, else an iterator over iterable
    """
    return iter(iterable) if iterable is not None else iter(tuple())


def call_if_nnone(clbl: tp.Optional[tp.Callable[..., V]], *args, **kwargs) -> tp.Optional[V]:
    """
    Call clbl with provided arguments, but only if clbl is not None.

    .. deprecated:: 2.24.2
        Use Optional(clbl)(*args, **kwargs) instead

    If it's None, then None will be returned. Else, the result of this callable will be returned.

    :param clbl: callable to call, or a None
    :param args: positional arguments to provide to clbl
    :param kwargs: keyword arguments to provide to clbl
    :return: return value or None
    """
    warnings.warn('This is deprecated, use Optional(clbl)() instead', DeprecationWarning)
    return clbl(*args, **kwargs) if clbl is not None else None


class Optional(Proxy[T]):
    """
    A wrapper for your classes, that does nothing if the object passed in is None.
    It will return an empty Optional in that case.

    Usage example:

    >>> may_be_none = None
    >>> Optional(may_be_none).cancel().result()

    So far operations supported:

    * calling
    * getattr
    * getitem/setitem/delitem
    * testing for truth
    * comparison (with nonexistent elements always comparing false)
    * membership test (with nonexistent elements always returning false)

    .. warning:: Returned objects via getattr and getitem are NOT wrapped in an
        Optional. You need to do it by hand or just file an issue. I'll add that when I
        need it.

    :param obj: object to wrap
    """

    def __init__(self, obj):
        super().__init__(obj)

    def __contains__(self, item) -> bool:
        me = getattr(self, '_Proxy__obj')
        if me is None:
            return False
        return item in me

    def __eq__(self, other) -> bool:
        me = getattr(self, '_Proxy__obj')
        if me is None:
            return False
        return other == me

    def __getattr__(self, item):
        obj = getattr(self, '_Proxy__obj')
        return EMPTY_OPTIONAL if obj is None else getattr(obj, item)

    def __call__(self, *args, **kwargs):
        obj = getattr(self, '_Proxy__obj')
        return EMPTY_OPTIONAL if obj is None else obj(*args, **kwargs)

    def __bool__(self):
        obj = getattr(self, '_Proxy__obj')
        return False if obj is None else bool(obj)

    def __getitem__(self, item):
        obj = getattr(self, '_Proxy__obj')
        return EMPTY_OPTIONAL if getattr(self, '_Proxy__obj') is None else obj[item]

    def __setitem__(self, key, value) -> None:
        obj = getattr(self, '_Proxy__obj')
        if obj is not None:
            obj.__setitem__(key, value)

    def __delitem__(self, key) -> None:
        obj = getattr(self, '_Proxy__obj')
        if obj is not None:
            del obj[key]

    def __len__(self) -> int:
        obj = getattr(self, '_Proxy__obj')
        return 0 if obj is None else len(obj)


EMPTY_OPTIONAL = Optional(None)


def extract_optional(v: tp.Union[T, Optional[T]]) -> T:
    """
    If v is an optional, extract the value that it wraps.
    If it is not, return v

    :param v: value to extract the value from
    :return: resulting value
    """
    return getattr(v, '_Proxy__obj') if isinstance(v, Optional) else v



import time
import typing as tp
from threading import Event

from satella.coding.misc import _BLANK
from satella.coding.typing import T
from satella.exceptions import WouldWaitMore


class DeferredValue(tp.Generic[T]):
    """
    A class that allows you to pass arguments that will be available later during runtime.

    Usage:

    >>> def thread1(value):
     >>>     print(value.value())

    >>> val = DeferredValue()
    >>> threading.Thread(target=thread1, args=(val, )).start()
    >>> time.sleep(10)
    >>> val.set_value(3)
    """

    __slots__ = 'val', 'lock'

    def __init__(self):
        self.lock = Event()
        self.val = _BLANK

    def set_value(self, va: T) -> None:
        """
        Set a value and wake up all the threads waiting on it.

        :param va: value to set
        :raises ValueError: value is already set
        """
        if self.val is not _BLANK:
            raise ValueError('Value curently set!')
        self.val = va
        self.lock.set()

    def result(self, timeout=None) -> T:
        """An alias for :meth:`~satella.coding.concurrent.DeferredValue.value`"""
        return self.value(timeout)

    def value(self, timeout: tp.Optional[float] = None) -> T:
        """
        Wait until value is available, and return it.

        :param timeout: number of seconds to wait. If None is given, this will take as long as necessary.
        :return: a value
        :raises WouldWaitMore: timeout was given and it has expired
        """
        if self.val is not _BLANK:
            return self.val
        tout = self.lock.wait(timeout)
        if not tout:
            raise WouldWaitMore()
        return self.val



from satella.coding.decorators import wraps


class EmptyContextManager:
    """
    A context manager that does nothing. Only to support conditional change of context managers,
    eg in such a way:

    >>> if tx_enabled:
    >>>     ctxt = transaction.atomic
    >>> else:
    >>>     ctxt = EmptyContextManager()
    >>> with ctxt:
    >>>     ...

    Note that it will accept any parameters, and then throw them on the ground.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


def wrap_callable_in_context_manager(clbl, ctxt_mgr, *my_args, **my_kwargs):
    """
    Wrap a callable in context manager.

    Roughly equivalent to:

    >>> def inner(*args, **kwargs):
    >>>     with ctxt_mgr(*my_args, **my_kwargs):
    >>>         return clbl(*args, **kwargs)
    >>> return inner

    To be used as:

    >>> clbl = wrap_callable_in_context_manager(lambda y: 5, tracing.start_new_span, 'New span')
    """
    @wraps(clbl)
    def inner(*args, **kwargs):
        with ctxt_mgr(*my_args, **my_kwargs):
            return clbl(*args, **kwargs)
    return inner



from __future__ import annotations

import threading
import typing as tp
import warnings

from satella.coding.typing import V

local = threading.local()

THEY_HATIN = object()


class Context:
    """
    New layer of environment. Can have it's own variables, or can hoist them onto the parent.

    .. warning:: This is considered experimental. I just haven't found out a good use case for it yet.
    """

    def __init__(self, parent: tp.Optional[Context] = None, **variables):
        warnings.warn('This is experimental', RuntimeWarning)
        self.parent = parent
        self.variables = variables
        self.bool = None

    def __str__(self):
        return str(id(self))

    def push_up(self, item: str, value=THEY_HATIN) -> None:
        """
        Advance current variable to the top of the card stack.

        :param item: variable name
        :param value: if not given, current value of given variable will be taken
        """
        if value is THEY_HATIN:
            var = self.variables.pop(item)
        else:
            var = value
        self.parent.variables[item] = var

    def __getattr__(self, item: str):
        if item in self.variables:
            v = self.variables[item]
            if v is not THEY_HATIN:
                return v
            raise AttributeError()
        if self.parent is not None:
            return getattr(self.parent, item)
        raise AttributeError()

    def __enter__(self):
        global local
        try:
            parent = local.thread_context
        except AttributeError:
            parent = None
        ctxt = Context(parent=parent)
        ctxt.parent = parent
        local.thread_context = ctxt
        return ctxt

    def __setattr__(self, key: str, value: V):
        """
        Set a value
        """
        if key in ('parent', 'variables', 'bool'):
            return super().__setattr__(key, value)
        else:
            self.variables[key] = value

    def __delattr__(self, item: str) -> None:
        self.variables[item] = THEY_HATIN

    def does_exist(self, val: str) -> bool:
        """
        Does a given value exist on stack for this call of function?
        """
        if val in self.variables:
            if self.variables[val] is THEY_HATIN:
                return False
            return True
        else:
            if self.parent is None:
                return False
            return self.parent.does_exist(val)

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        if self.parent is None:
            try:
                del local.thread_context
            except AttributeError:
                pass
        else:
            local.thread_context = self.parent
        return False

    @staticmethod
    def get() -> Context:
        """
        Return a local context for this thread
        """
        global local
        try:
            return local.thread_context
        except AttributeError:
            ctxt = Context()
            local.thread_context = ctxt
            return ctxt



import typing as tp

from satella.coding.typing import V


def list_values_to_indices(lst: tp.List[V]) -> tp.Dict[V, int]:
    """
    Transform a list of entries into a dict mapping where given entry can be found.

    Example:

    >>> a = ['abc', 'def', 'ghi']
    >>> b = list_values_to_indices(a)
    >>> assert b == {'abc': 0, 'def': 1, 'ghi': 2}

    :param lst: list to process. Take care for the list to be composed of unique entries.
    :raises ValueError: item was found more than once
    """
    result = {}
    for idx, val in enumerate(lst):
        if val in result:
            raise ValueError(f'Entry {val} found more than once!')
        result[val] = idx
    return result



import unittest
import warnings

from satella.warnings import ExperimentalWarning


class TestWarning(unittest.TestCase):
    def test_experimental_warning(self):
        warnings.warn('Experimental', ExperimentalWarning)



import unittest

from satella.debug import TaintedObject, TaintingEnvironment


class TestTainting(unittest.TestCase):
    def test_declare(self):
        self.assertRaises(RuntimeError, lambda: TaintedObject(5))

        with TaintingEnvironment() as env:
            a = 5
            b = TaintedObject(a)
            c = TaintedObject(b)
            self.assertIs(b, c)

            def taint_me(a, b, c):
                return a+b+c
            d = taint_me(a, b, c)
            self.assertIsInstance(d, TaintedObject)
            self.assertEqual(d, 15)
            self.assertIn(15, list(env.get_tainted_variables()))



from __future__ import annotations

import inspect
import typing as tp


# Taken from https://stackoverflow.com/questions/28237955/same-name-for-classmethod-and-
# instancemethod


class class_or_instancemethod(classmethod):
    """
    A decorator to make your methods both classmethods (they will receive an instance of type
    as their first argument) or normal methods (they will receive an instance of their type).

    Use like:

    >>> class MyClass:
    >>>     @class_or_instancemethod
    >>>     def method(self_or_cls):
    >>>         if isinstance(self_or_cls, MyClass):
    >>>             # method code
    >>>         else:
    >>>             # classmethod code
    """

    def __get__(self, instance, type_):
        descr_get = super().__get__ if instance is None else self.__func__.__get__
        return descr_get(instance, type_)


class TypeSignature(inspect.Signature):
    """
    A type signature.

    You can compare signatures:

    >>> def a(y: object):
    >>>     pass
    >>> def b(y: int):
    >>>     pass
    >>> TypeSignature.from_fun(a) < TypeSignature(b)
    """
    __slots__ = ()

    # pylint: disable=protected-access
    def __init__(self, t_sign: inspect.Signature):
        """
        :param t_sign: a inspect.Signature
        """
        self._return_annotation = t_sign._return_annotation
        self._parameters = t_sign._parameters

    @staticmethod
    def from_fun(fun) -> TypeSignature:
        """Return a type signature from a function"""
        return TypeSignature(inspect.Signature.from_callable(fun))

    def can_be_called_with_args(self, *args, **kwargs) -> bool:
        """Can this type signature be called with following arguments?"""
        called = self._bind(*args, **kwargs)

        # pylint: disable=protected-access
        return all(issubclass(self.signature.parameters.get(arg_name, NONEARGS)._annotation, arg_value)
                   for arg_name, arg_value in called.items())

    def is_more_generic_than(self, b: TypeSignature) -> bool:
        """Is this type signature more generic than an other?"""
        if self == {}:
            for key in self:
                key1 = self[key]
                key2 = b.get(key, None)
                if key2 is None:
                    return key2 == {}

                if key2.is_more_generic_than(key1):
                    return False
        return True

    def __lt__(self, other: TypeSignature) -> bool:
        return self.is_more_generic_than(other)

    def matches(self, *args, **kwargs) -> bool:
        """
        Does this invocation match this signature?
        """
        bound_args = self.bind(*args, **kwargs)
        bound_args.apply_defaults()
        for param_name, param_value in bound_args.arguments.items():
            if not isinstance(param_value, self._parameters[param_name].annotation):
                return False
        return True


class overload:
    """
    A class used for method overloading.

    .. warning::

       This feature is scheduled for an overhaul and may not work as promised. Keep that in mind.

    Note that methods can be only overloaded by their positional, or positional-and-keyword
    arguments. Overload distinguishment will be done at the level of positional arguments only.

    Note that typing checks will be done via isinstance().

    Use like this:

    >>> @overload
    >>> def what_type(x: str):
    >>>     print('String')
    >>> @what_type.overload
    >>> def what_type(x: int):
    >>>     print('Int')
    >>> what_type(5)
    >>> what_type('string')

    Note that this instance's __wrapped__ will refer to the first function.
    TypeError will be called if no signatures match arguments.
    """

    def __init__(self, fun: tp.Callable):
        self.type_signatures_to_functions = {TypeSignature.from_fun(fun): fun}
        if hasattr(fun, '__doc__'):
            self.__doc__ = fun.__doc__
        self.__wrapped__ = fun

    @property
    def all_functions(self) -> tp.Iterable[object]:
        """Return a list of all functions registered within this overload"""
        return self.type_signatures_to_functions.values()

    def overload(self, fun):
        """
        :raises ValueError: this signature already has an overload
        """
        sign = TypeSignature.from_fun(fun)
        if sign in self.type_signatures_to_functions:
            raise TypeError('Method of this signature is already overloaded with %s' % (fun,))
        self.type_signatures_to_functions[sign] = fun
        return self

    def __call__(self, *args, **kwargs):
        """
        Call one of the overloaded functions.

        :raises TypeError: no type signature given
        """
        matching = []
        for sign, fun in self.type_signatures_to_functions.items():
            if sign.matches(*args, **kwargs):
                matching.append((sign, fun))
        matching.sort()     # This sorting should result in most precise class at the end of the list
        if not matching:
            raise TypeError('No matching entries!')
        else:
            return matching[-1][1](*args, **kwargs)  # call the most specific function you could find



import abc
import logging
import queue
import sys
import typing as tp
import warnings

from ..concurrent import Monitor
from ..misc import Closeable
from ..recast_exceptions import silence_excs
from ..typing import T

logger = logging.getLogger(__name__)


class CPManager(Monitor, Closeable, tp.Generic[T], metaclass=abc.ABCMeta):
    """
    A thread-safe no-hassle connection-pool manager.

    Extend this class to build your own connection pool managers.

    This supports automatic connection recycling, connection will be cycled each
    max_cycle_no takings and deposits.

    Note that you have to overload :meth:`~satella.coding.resources.CPManager.teardown_connection`
    and :meth:`~satella.coding.resources.CPManager.create_connection`.

    You obtain a connection by using :meth:`~satella.coding.resources.CPManager.acquire_connection`.
    If it fails you should mark it as such using
    :meth:`~satella.coding.resources.CPManager.fail_connection`.
    In all cases you have to return it using
    :meth:`~satella.coding.resources.CPManager.release_connection`.

    :param max_number: maximum number of connections
    :param max_cycle_no: maximum number of get/put connection cycles.
    :ivar max_number: maximum amount of connections. Can be changed during runtime

    .. warning:: May not work under PyPy for reasons having to do with id's semantics.
        A RuntimeWarning will be issued when not running under CPython.
    """

    def __init__(self, max_number: int, max_cycle_no: int):
        Closeable.__init__(self)
        Monitor.__init__(self)
        if sys.implementation.name != 'cpython':
            warnings.warn(f'This may run bad on {sys.implementation.name}', RuntimeWarning)
        self.connections = queue.Queue(max_number)
        self.max_number = max_number
        self.max_cycle_no = max_cycle_no
        self.id_to_times = {}  # type: tp.Dict[int, int]
        self.terminating = False

    def close(self) -> None:
        if super().close():
            self.terminating = True
            self.invalidate()

    @Monitor.synchronized
    def invalidate(self) -> None:
        """
        Close all connections. Connections have to be released first. Object is ready for use after this
        """
        while self.connections.qsize() > 0:
            conn = self.connections.get()
            self.teardown_connection(conn)
            del self.id_to_times[id(conn)]

    def acquire_connection(self) -> T:
        """
        Either acquire a new connection, or just establish it in the background

        :return: a new connection:
        :raises RuntimeError: CPManager is terminating!
        """
        if self.terminating:
            raise RuntimeError('CPManager is terminating')
        try:
            conn = self.connections.get(False)
        except queue.Empty:
            while True:
                with silence_excs(queue.Empty), Monitor.acquire(self):
                    if self.connections.qsize() >= self.max_number:
                        conn = self.connections.get(False, 5)
                        break
                    else:
                        conn = self.create_connection()
                        break
        with Monitor.acquire(self):
            obj_id = id(conn)
            self.id_to_times[obj_id] = self.id_to_times.get(obj_id, 0) + 1
            return conn

    def release_connection(self, connection: T) -> None:
        """
        Release a connection

        :param connection: connection to release
        """
        obj_id = id(connection)
        if self.id_to_times[obj_id] >= self.max_cycle_no:
            self._kill_connection(connection)
        else:
            try:
                self.connections.put(connection, False)
            except queue.Full:
                self._kill_connection(connection)

    def _kill_connection(self, connection):
        obj_id = id(connection)
        with Monitor.acquire(self):
            del self.id_to_times[obj_id]

        self.teardown_connection(connection)

    @Monitor.synchronized
    def fail_connection(self, connection: T) -> None:
        """
        Signal that a given connection has been failed

        :param connection: connection to fail
        """
        self.id_to_times[id(connection)] = self.max_cycle_no

    @abc.abstractmethod
    def teardown_connection(self, connection: T) -> None:
        """
        Close the connection.

        Is safe to block.

        :param connection: connection to tear down
        """

    @abc.abstractmethod
    def create_connection(self) -> T:
        """
        Create a new connection.

        Is safe to block.

        :return: a new connection instance
        """



from __future__ import annotations

import dis
import gc
import sys
import threading
import typing as tp
import warnings

from satella.coding.typing import T
from .tainteds import TaintedObject, access_tainted, taint

local = threading.local()

RET_VAL = dis.opmap['RETURN_VALUE']


class TaintingEnvironment:
    """
    A round of tainting. Taints will be invalidated at the end.
    Basically any Python opcode that relates to tainting will
    be tainted itself.

    As an added extra, function values will be tainted as well if at least one of the parameters has been tainted

    Use like this:

    >>> with TaintingEnvironment() as env:
    >>>     a = taint(a)
    >>>     b = a + 5
    >>>     ...

    .. warning:: Using environment tainting will slow down your Python scripts since they install a per opcode handler.

    .. warning:: Using functions that take at least one tainted argument is supposed to return a tainted result,
        but I couldn't get that shipped in time at this involved some serious tweaking with the Python bytecode.
    """
    __slots__ = 'enabled', 'old_python'

    def __init__(self):
        vi = sys.version_info
        self.old_python = False
        if vi.major >= 3 and vi.minor < 7:
            warnings.warn(
                'Due to an old Python being used, altering function return values to be tainted will not work',
                UserWarning)
            self.old_python = True
        elif sys.implementation.name != 'cpython':
            warnings.warn(
                'You are not using CPython. Since this library will eventually involve some tweaking with the Python '
                'bytecode, if your runtime does not provide this level of compatibility, it will not work.',
                UserWarning)

        self.enabled = False

    def __enter__(self) -> TaintingEnvironment:
        """
        Register itself as the current tainting environment

        :raises RuntimeError: there is already a tainting session in progress
        """
        self.enabled = True
        global local
        if hasattr(local, 'satella_tainting'):
            raise RuntimeError('Other tainting session in progress')
        local.satella_tainting = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del local.satella_tainting
        self.enabled = False
        return False

    @staticmethod
    def get_session_in_progress() -> TaintingEnvironment:
        """
        Obtain current tainting session in progress

        :raises RuntimeError: no tainting session started yet
        """
        if not hasattr(local, 'satella_tainting'):
            raise RuntimeError('You require a session in progress to do that')
        return local.satella_tainting

    def get_tainted_variables(self) -> tp.Iterator[T]:
        """
        Return all, of the tainted variables
        """
        for obj in gc.get_referrers(self):
            if not isinstance(obj, TaintedObject):
                continue
            yield access_tainted(obj)



import unittest
from satella.random import shuffle_together, random_binary, random_word
import string


class TestRandom(unittest.TestCase):

    def test_random_word(self):
        rand_word = random_word(5)
        self.assertEqual(len(rand_word), 5)
        self.assertTrue(all(y in string.ascii_lowercase for y in rand_word))

    def test_random_binary(self):
        self.assertFalse(random_binary(0))
        self.assertEqual(len(random_binary(10)), 10)

    def test_random(self):
        a = [(1, 2, 3), ('a', 'b', 'c')]
        b = zip(*shuffle_together(*a))
        self.assertEqual(set(b), {(1, 'a'), (2, 'b'), (3, 'c')})

        self.assertEqual(shuffle_together(), [])





import unittest

from satella.coding import overload, TypeSignature


class TestOverloading(unittest.TestCase):
    def test_type_signature(self):
        def a(a: object):
            pass

        def b(a: int):
            pass

        self.assertLess(TypeSignature.from_fun(a), TypeSignature.from_fun(b))

    def test_something(self):
        @overload
        def fun(i: int):
            self.assertIsInstance(i, int)

        @fun.overload
        def fun(i: str):
            self.assertIsInstance(i, str)

        fun(2)
        self.assertRaises(TypeError, lambda: fun(2.5))
        fun('test')

    def test_something_2(self):
        class A:
            pass

        class B(A):
            pass

        @overload
        def fun(a: B):
            self.assertEqual(type(a), B)

        @fun.overload
        def fun(a: A):
            self.assertEqual(type(a), A)

        fun(A())
        fun(B())



import typing as tp
import time
from concurrent.futures import wait, ThreadPoolExecutor

from satella.coding.concurrent.atomic import AtomicNumber
from satella.coding.concurrent.futures import ExecutorWrapper
from satella.coding.concurrent.thread import Condition
from satella.exceptions import WouldWaitMore
from satella.time.measure import measure


def _while_sync_threadpool(tpe, max_wait, measurement, futures):
    while tpe._work_queue.qsize() > 0:      # pylint: disable=protected-access
        if max_wait is not None:
            if measurement() > max_wait:
                for future in futures:
                    future.cancel()
                raise WouldWaitMore('timeout exceeded')
        time.sleep(0.5)


def _max_wait_atm_n(measurement, max_wait, atm_n):
    while measurement() < max_wait:
        try:
            atm_n.wait_until_equal(0, 1)
            return
        except WouldWaitMore:
            continue
    raise WouldWaitMore('timeout exceeded')


def sync_threadpool(tpe: tp.Union[ExecutorWrapper, ThreadPoolExecutor],
                    max_wait: tp.Optional[float] = None) -> None:
    """
    Make sure that every thread of given thread pool executor is done processing
    jobs scheduled until this moment.

    Make sure that other tasks do not submit anything to this thread pool executor.

    :param tpe: thread pool executor to sync. Can be also a ExecutorWrapper.
    :param max_wait: maximum time to wait. Default, None, means wait forever
    :raises WouldWaitMore: timeout exceeded. Raised only when max_wait is not None.
    """
    if isinstance(tpe, ExecutorWrapper):
        return sync_threadpool(tpe.executor, max_wait=max_wait)

    assert isinstance(tpe, ThreadPoolExecutor), 'Must be a ThreadPoolExecutor!'

    with measure(timeout=max_wait) as measurement:
        workers = tpe._max_workers      # pylint: disable=protected-access
        atm_n = AtomicNumber(workers)
        cond = Condition()

        def decrease_atm():
            nonlocal atm_n
            atm_n -= 1
            cond.wait()

        futures = [tpe.submit(decrease_atm) for _ in range(workers)]

        # wait for all currently scheduled jobs to be picked up
        _while_sync_threadpool(tpe, max_wait, measurement, futures)

        if max_wait is None:
            atm_n.wait_until_equal(0)
        else:
            _max_wait_atm_n(measurement, max_wait, atm_n)
        cond.notify_all()
        wait(futures)



import signal
import unittest
import time
import multiprocessing
import os
import sys

from satella.coding.concurrent import call_in_separate_thread
from satella.time import measure, time_as_int, time_ms, sleep, ExponentialBackoff, \
    parse_time_string
from concurrent.futures import Future
from satella.warnings import mark_temporarily_disabled


class TestTime(unittest.TestCase):

    def test_mark_temp_disabled(self):

        @mark_temporarily_disabled('Skipped due to skip')
        def dupa(a):
            raise ValueError()

        self.assertRaises(NotImplementedError, dupa)

    def test_measure_as_method(self):
        self2 = self

        class MyClass:
            def __init__(self):
                self.a = 1

            @measure()
            def function(self, v, b):
                time.sleep(1)
                self.a += b
                self2.assertGreaterEqual(v(), 1)

        d = MyClass()
        with measure() as m:
            d.function(2)
        self.assertGreaterEqual(m(), 1)
        self.assertEqual(d.a, 3)

    def test_exponential_backoff_grace(self):
        eb = ExponentialBackoff(grace_amount=2)
        self.assertTrue(eb.available)
        eb.failed()
        self.assertTrue(eb.available)
        eb.failed()
        self.assertTrue(eb.available)
        eb.failed()
        self.assertFalse(eb.available)

    def test_exponential_backoff_earlier_wakeup(self):
        eb = ExponentialBackoff(start=5, limit=30)

        @call_in_separate_thread()
        def do_test():
            with measure() as m:
                eb.wait_until_available(2.5)
                self.assertTrue(eb.available)
            self.assertLessEqual(m(), 1.5)

        eb.failed()
        do_test()
        time.sleep(1)
        eb.success()

    def test_exponential_backoff_launch_immediate(self):
        eb = ExponentialBackoff(start=2, limit=30)
        i = 1
        a = 0
        with measure() as m:
            @eb.launch(ValueError, immediate=True)
            def do_action():
                nonlocal i, a
                a += 1
                if i == 4:
                    return
                else:
                    i += 1
                    raise ValueError()
        self.assertGreaterEqual(m(), 2)
        self.assertGreaterEqual(a, 2)

    def test_exponential_backoff_launch(self):
        eb = ExponentialBackoff(start=2, limit=30)
        i = 1
        a = 0
        @eb.launch(ValueError)
        def do_action():
            nonlocal i, a
            a += 1
            if i == 4:
                return
            else:
                i += 1
                raise ValueError()

        with measure() as m:
            do_action()
        self.assertGreaterEqual(m(), 2)
        self.assertGreaterEqual(a, 2)

    def test_exponential_backoff_waiting_for_service_healthy(self):
        eb = ExponentialBackoff(start=2, limit=30)
        self.assertTrue(eb.ready_for_next_check)

        eb.failed()
        self.assertFalse(eb.available)
        self.assertFalse(eb.ready_for_next_check)
        with measure() as m:
            eb.wait_until_available()
        self.assertTrue(eb.ready_for_next_check)
        self.assertGreaterEqual(m(), 1.8)
        eb.success()
        self.assertTrue(eb.ready_for_next_check)
        self.assertTrue(eb.available)
        eb.failed()
        self.assertFalse(eb.available)
        with measure() as m:
            eb.sleep()
        self.assertGreaterEqual(m(), 2)

    def test_parse_time_string(self):
        self.assertEqual(parse_time_string('30m'), 30 * 60)
        self.assertEqual(parse_time_string('30h'), 30*60*60)
        self.assertEqual(parse_time_string('30w'), 30 * 7 * 24 * 60 * 60)
        self.assertEqual(parse_time_string(2), 2)
        self.assertEqual(parse_time_string(2.0), 2.0)

    def test_exponential_backoff(self):
        with measure() as measurement:
            eb = ExponentialBackoff()
            eb.failed()
            eb.sleep()
            eb.failed()
            eb.sleep()
        self.assertGreaterEqual(measurement(), 3)

    def test_measure(self):
        with measure(timeout=0.5) as measurement:
            self.assertFalse(measurement.timeouted)
            time.sleep(1)
            self.assertTrue(measurement.timeouted)

    def test_measure_reset_and_stop(self):
        m1 = measure(create_stopped=True)
        m1.start()
        time.sleep(0.5)
        self.assertFalse(m1.stopped)
        self.assertGreaterEqual(m1(), 0.5)
        m1.reset_and_stop()
        self.assertTrue(m1.stopped)
        self.assertEqual(m1(), 0)
        m1.start()
        time.sleep(0.5)
        self.assertGreaterEqual(m1(), 0.5)

    @unittest.skipIf('win' in sys.platform, 'Needs POSIX to run')
    @unittest.skipIf(sys.platform != 'cpython', 'Seems to have some trouble cooperating with PyPy')
    def test_sleep(self):
        sleep(-2)

        def runner():
            time.sleep(1)
            os.kill(os.getppid(), signal.SIGINT)

        multiprocessing.Process(target=runner).start()

        try:
            with measure() as measurement:
                sleep(3)
            self.assertGreaterEqual(measurement(), 3)
        finally:
            os.wait()

        multiprocessing.Process(target=runner).start()

        try:
            with measure() as measurement:
                sleep(3, True)
            self.assertLessEqual(measurement(), 3)
        finally:
            os.wait()

    def test_times(self):
        self.assertIsInstance(time_as_int(), int)
        self.assertIsInstance(time_ms(), int)

    def test_measure_future(self):
        future = Future()
        measurement = measure(future)
        future.set_running_or_notify_cancel()
        time.sleep(1)
        future.set_result(2)
        self.assertEqual(future.result(), 2)
        self.assertGreaterEqual(measurement(), 1)

    def test_measure(self):
        with measure() as measurement:
            time.sleep(0.5)

        self.assertGreaterEqual(measurement(), 0.5)

    def test_measure_decorator(self):
        @measure()
        def measured(measurement):
            time.sleep(0.5)
            self.assertGreaterEqual(measurement(), 0.5)
        measured()

    def test_measure_method_decorator(self):
        class Test:
            @measure()
            def try_measure(self, measurement):
                time.sleep(0.5)
                return measurement()

        for i in range(3):
            self.assertTrue(0.5 <= Test().try_measure() <= 0.7)

    def test_measure_method_decorator_stopped(self):
        class Test:
            @measure(create_stopped=True)
            def try_measure(self, measurement):
                time.sleep(0.3)
                measurement.start()
                time.sleep(0.5)
                return measurement()

        for i in range(3):
            self.assertTrue(0.5 <= Test().try_measure() <= 0.7)

    def test_resuming(self):
        a = measure()
        self.assertRaises(TypeError, lambda: a.start())
        time.sleep(0.5)
        self.assertGreaterEqual(a(), 0.5)
        a.stop()
        time.sleep(0.5)
        self.assertLess(a(), 0.7)
        a.start()
        time.sleep(0.5)
        self.assertLess(a(), 1.2)
        a.stop()
        self.assertRaises(TypeError, lambda: a.stop())
        a.reset_and_start()
        time.sleep(0.5)
        self.assertGreaterEqual(a(), 0.5)
        self.assertLess(a(), 0.7)



from __future__ import annotations

import functools
import typing as tp

from satella.coding.typing import T


def might_accept_tainted(taint_result: bool = False):
    """
    For methods that might accept a tainted value during execution
    
    This both unpacks your first argument if is was a TaintedObject, you'll receive it's value
    :param taint_result: result will be tainted
    """

    def outer(fun):
        @functools.wraps(fun)
        def inner(self, *args):
            if len(args) > 0:
                args = access_tainted(args[0]), *args[1:]
            p = fun(self, *args)
            if taint_result:
                return taint(p)
            else:
                return p

        return inner

    return outer


class TaintedObject(tp.Generic[T]):
    __slots__ = '__environment', '__v'

    def __new__(cls, v):
        if isinstance(v, TaintedObject):
            return v
        return object.__new__(cls)

    def __init__(self, v: T):
        from .environment import TaintingEnvironment
        if isinstance(v, TaintedObject):
            # all is already set correctly for this object and it will be returned
            return
        self.__environment = TaintingEnvironment.get_session_in_progress()
        self.__v = v
        if isinstance(v, tp.MutableSequence):
            for i in range(len(v)):
                self.__v[i] = taint(v[i])
        else:
            try:
                if isinstance(v, tp.MutableMapping):
                    dict_to_transform = v
                else:
                    dict_to_transform = v.__dict__
                for key in dict_to_transform:
                    dict_to_transform[key] = taint(dict_to_transform[key])
            except AttributeError:
                pass  # must have been a primitive type

    def __setattr__(self, key: str, value: tp.Any):
        if key == '_TaintedObject__environment' or key == '_TaintedObject__v':
            super().__setattr__(key, value)
        else:
            super().__setattr__(key, taint(value))

    def __iter__(self) -> tp.Iterator[TaintedObject[T]]:
        for item in self.__v:
            yield taint(item)

    def __next__(self) -> TaintedObject[T]:
        return taint(next(self.__v))

    @might_accept_tainted()
    def __str__(self) -> str:  # it must be a str otherwise Python will complain
        return str(self.__v)

    def __call__(self, *args, **kwargs):
        return taint(self.__v(*args, **kwargs))

    @might_accept_tainted(taint_result=True)
    def __repr__(self) -> str:
        return '%s TAINTED' % str(self.__v)

    @might_accept_tainted(taint_result=True)
    def __getattr__(self, item: str) -> TaintedObject[T]:
        return getattr(self.__v, item)

    def __len__(self) -> int:  # This has to return an int otherwise Python will break
        return len(self.__v)

    def __bool__(self) -> bool:  # This has to return a bool otherwise Python will break
        return bool(self.__v)

    def __int__(self) -> int:
        return int(self.__v)

    def __float__(self) -> float:
        return float(self.__v)

    def __enter__(self):
        return taint(self.__v.__enter__())

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.__v.__exit__(exc_type, exc_val, exc_tb)

    @might_accept_tainted(taint_result=True)
    def __eq__(self, other) -> bool:
        return self.__v == other

    @might_accept_tainted(taint_result=True)
    def __gt__(self, other) -> bool:
        return self.__v > other

    @might_accept_tainted(taint_result=True)
    def __ge__(self, other) -> bool:
        return self.__v >= other

    @might_accept_tainted(taint_result=True)
    def __lt__(self, other) -> bool:
        return self.__v < other

    @might_accept_tainted(taint_result=True)
    def __le__(self, other) -> bool:
        return self.__v <= other

    @might_accept_tainted(taint_result=True)
    def __add__(self, other) -> TaintedObject:
        return self.__v + other

    @might_accept_tainted(taint_result=True)
    def __radd__(self, other) -> TaintedObject:
        return other + self.__v

    @might_accept_tainted()
    def __iadd__(self, other) -> TaintedObject:
        self.__v += other
        return self

    @might_accept_tainted(taint_result=True)
    def __sub__(self, other) -> TaintedObject:
        return self.__v - other

    @might_accept_tainted()
    def __isub__(self, other) -> TaintedObject:
        self.__v -= other
        return self

    @might_accept_tainted(taint_result=True)
    def __mul__(self, other) -> TaintedObject:
        return self.__v * other

    @might_accept_tainted()
    def __imul__(self, other) -> TaintedObject:
        self.__v *= other
        return self

    def __dir__(self) -> tp.Iterable[str]:
        return dir(self.__v)

    def __hash__(self) -> int:
        return hash(self.__v)

    def __getitem__(self, item: int) -> TaintedObject[T]:
        return taint(self.__v[item])

    def __setitem__(self, key, value: T) -> None:
        self.__v[key] = taint(value)

    def __delitem__(self, key: str) -> None:
        del self.__v[key]


def access_tainted(v: tp.Union[T, TaintedObject[T]]) -> T:
    """
    If v is tainted, this will extract it's value.

    If it is not, v will be returned
    """
    if not isinstance(v, TaintedObject):
        return v
    return getattr(v, '_TaintedObject__v')


def taint(v: T) -> TaintedObject[T]:
    """
    Taints the object if necessary. If already tainted will leave it as is

    :raises RuntimeError: no tainting session in progress
    """
    return v if isinstance(v, TaintedObject) else TaintedObject(v)



import logging
import os
import platform
import signal
import sys

from satella.instrumentation import install_dump_frames_on
from satella.instrumentation.memory import MemoryPressureManager, CustomCondition, All, Any, \
    dump_memory_on, get_size
import time
import unittest
logger = logging.getLogger(__name__)


class OnDemandCondition(CustomCondition):
    def __init__(self):
        self.value = False
        super().__init__(lambda: self.value)

    def can_fire(self, *args) -> bool:
        return self.value


class TestMemory(unittest.TestCase):
    @unittest.skipIf(platform.python_implementation(), 'This will not work on PyPy')
    def test_get_size_dict(self):
        a = {'aba': 'aba'}

        class Aba:
            def __init__(self):
                self.aba = 'aba'

        self.assertGreater(get_size(a), 6)
        self.assertGreater(get_size(Aba()), 6)

    @unittest.skipIf(sys.platform == 'win32', 'testable only on unices')
    def test_install_dump_on(self):
        install_dump_frames_on(signal.SIGUSR1)
        os.kill(os.getpid(), signal.SIGUSR1)

    @unittest.skipIf(platform.python_implementation() == 'PyPy', 'does not work on PyPy')
    def test_get_size(self):
        a = 'a' * 1024
        self.assertGreaterEqual(get_size(a), 1024)

    @unittest.skipIf(platform.python_implementation() != 'PyPy', 'requires PyPy')
    def test_get_size_runtime_error(self):
        a = 'a' * 1024
        self.assertRaises(RuntimeError, lambda: get_size(a))

    def test_dump_memory(self):
        dump_memory_on()

    def test_memory(self):
        odc = OnDemandCondition()

        a = {'memory': False,
             'calls': 0,
             'improved': False,
             'times_entered_1': 0,
             'level_2_engaged': False,
             'level_2_confirmed': False,
             'cancelled': 0,
             'mem_normal': 0}

        class ObjectToCleanup:
            def __init__(self):
                self.cleaned_up = False

            def cleanup(self):
                self.cleaned_up = True

        obj1 = ObjectToCleanup()
        obj2 = ObjectToCleanup()

        cc = CustomCondition(lambda: a['level_2_engaged'])

        MemoryPressureManager(None, [odc, All(cc, Any(cc, cc))], 2)
        MemoryPressureManager.cleanup_on_entered(1, obj1)
        MemoryPressureManager.cleanup_on_entered(2, obj2)

        def memory_normal():
            nonlocal a
            a['mem_normal'] += 1

        def cancel():
            nonlocal a
            a['cancelled'] += 1
        MemoryPressureManager.register_on_memory_normal(memory_normal)

        cc = MemoryPressureManager.register_on_entered_severity(1)(cancel)

        @MemoryPressureManager.register_on_entered_severity(2)
        def call_on_level_2():
            a['level_2_confirmed'] = True

        @MemoryPressureManager.register_on_remaining_in_severity(1)
        def call_on_memory_still():
            a['calls'] += 1

        @MemoryPressureManager.register_on_entered_severity(1)
        def call_on_no_memory():
            a['memory'] = True
            a['times_entered_1'] += 1

        @MemoryPressureManager.register_on_left_severity(1)
        def call_improved():
            a['improved'] = True

        self.assertFalse(a['memory'])
        self.assertFalse(a['improved'])
        self.assertEqual(a['mem_normal'], 0)
        time.sleep(3)
        odc.value = True
        time.sleep(5)
        self.assertEqual(a['cancelled'], 1)
        cc.cancel()
        self.assertTrue(a['memory'])
        self.assertFalse(a['improved'])
        self.assertGreater(a['calls'], 0)
        self.assertTrue(obj1.cleaned_up)
        self.assertEqual(a['times_entered_1'], 1)
        del obj1
        odc.value = False
        time.sleep(3)
        self.assertTrue(a['improved'])
        self.assertEqual(a['times_entered_1'], 1)
        self.assertTrue(a['memory'])
        self.assertEqual(a['mem_normal'], 1)
        a['level_2_engaged'] = True
        time.sleep(3)
        self.assertLessEqual(len(MemoryPressureManager().objects_to_cleanup_on_entered[1]), 1)
        self.assertEqual(MemoryPressureManager().severity_level, 2)
        self.assertEqual(a['cancelled'], 1)
        self.assertEqual(a['times_entered_1'], 2)
        self.assertTrue(a['level_2_confirmed'])
        self.assertEqual(a['mem_normal'], 1)
        self.assertTrue(obj2.cleaned_up)



from .collector import calculate_occupancy_factor, sleep_cpu_aware, CPUTimeManager, get_own_cpu_usage, pCPUtimes
from .concurrency import CPUTimeAwareIntervalTerminableThread

__all__ = ['calculate_occupancy_factor', 'sleep_cpu_aware', 'pCPUtimes', 'get_own_cpu_usage',
           'CPUTimeAwareIntervalTerminableThread',
           'CPUTimeManager']



import inspect
import logging

logger = logging.getLogger(__name__)


class transaction:
    """
    A context manager for wrapping a transaction and getting a cursor from the Python DB API 2.

    Use it as a context manager. commit and rollback will be automatically called for you.

    Use like:

    >>> with transaction(conn) as cur:
    >>>   cur.execute('DROP DATABASE')

    Leaving the context manager will automatically close the cursor for you.

    :param connection_or_getter: the connection object to use, or a callable, that called with
        zero arguments will provide us with a connection
    :param close_the_connection_after: whether the connection should be closed after use
    :param log_exception: whether to log an exception if it happens
    """

    def __init__(self, connection_or_getter, close_the_connection_after: bool = False,
                 log_exception: bool = True):
        self._connection = connection_or_getter
        self.close_the_connection_after = close_the_connection_after
        self.cursor = None
        self.log_exception = log_exception

    def __enter__(self):
        self.cursor = self.connection.cursor()
        return self.cursor

    @property
    def connection(self):
        if inspect.isfunction(self._connection):
            conn = self._connection()
            self._connection = conn
            return conn
        else:
            return self._connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is None:
            self.connection.commit()
        else:
            self.connection.rollback()

        if exc_val is not None and self.log_exception:
            logger.error('Exception occurred of type %s', exc_type, exc_info=exc_val)

        self.cursor.close()

        if self.close_the_connection_after:
            self.connection.close()

        return False



import time
import unittest
from concurrent.futures import Future

from satella.coding.concurrent import call_in_separate_thread
from satella.coding.resources import CPManager


class TestResources(unittest.TestCase):
    def test_cp_manager(self):

        class Connection:
            total_connections = 0

            def __init__(self):
                self.i = 0
                Connection.total_connections += 1
                self.id = Connection.total_connections
                self.value_error_emitted = False
                self.closed = False

            def do(self):
                if self.closed:
                    raise RuntimeError('Connection closed')
                if self.value_error_emitted:
                    raise RuntimeError('do called despite raising ValueError earlier')
                self.i += 1
                if self.i == 3:
                    self.value_error_emitted = True
                    raise ValueError()
                elif self.i > 3:
                    raise RuntimeError('called too often')

            def close(self):
                self.closed = True

        class InheritCPManager(CPManager):
            def __init__(self, *args):
                super().__init__(*args)
                self.already_acquired = set()

            def create_connection(self) -> Connection:
                return Connection()

            def acquire_connection(self):
                v = super().acquire_connection()
                print(f'Acquiring connection {v.id}')
                if v.id in self.already_acquired:
                    raise RuntimeError('Reacquiring an acquired connection')
                self.already_acquired.add(v.id)
                return v

            def release_connection(self, connection) -> None:
                print(f'Releasing connection {connection.id}')
                self.already_acquired.remove(connection.id)
                super().release_connection(connection)

            def teardown_connection(self, connection) -> None:
                connection.close()

        cp = InheritCPManager(5, 3)

        conns = [cp.acquire_connection() for _ in range(5)]

        @call_in_separate_thread(daemon=True)
        def do_call():
            for _ in range(10):
                conn = cp.acquire_connection()
                time.sleep(1)
                try:
                    conn.do()
                except ValueError:
                    cp.fail_connection(conn)

                cp.release_connection(conn)

        ret = do_call()  # type: Future
        ret2 = do_call()
        cp.release_connection(conns.pop())
        cp.release_connection(conns.pop())
        ret.result(timeout=20)
        ret2.result(timeout=20)
        for i in range(20):
            conn = cp.acquire_connection()
            try:
                conn.do()
            except ValueError:
                cp.fail_connection(conn)
            cp.release_connection(conn)

        while conns:
            cp.release_connection(conns.pop())
        del cp



import unittest

from satella.dao import Loadable, must_be_loaded


class TestDAO(unittest.TestCase):
    def test_something(self):
        class Load(Loadable):
            def __init__(self, load_lazy=False):
                super().__init__(load_lazy=load_lazy)

            @must_be_loaded
            def method_accessed(self):
                assert self._loaded

            def refresh(self, load_from=None) -> None:
                super().refresh(load_from=load_from)

        l = Load()
        l.method_accessed()



import enum
import json
import typing as tp
import unittest

from satella.json import JSONAble, json_encode, read_json_from_file, write_json_to_file, \
    write_json_to_file_if_different, JSONEncoder


class TestJson(unittest.TestCase):

    def test_json_encoder_enums(self):
        enc = JSONEncoder()

        class Enum(enum.Enum):
            A = 'test'
        b = enc.encode(Enum.A)
        self.assertEqual(b, '"test"')

        write_json_to_file('test3.json', Enum.A)

    def test_write_json_to_file_if_different(self):
        d = {'test': 4}
        self.assertTrue(write_json_to_file_if_different('test2.json', d))
        self.assertFalse(write_json_to_file_if_different('test2.json', d))
        d = {'test': 5}
        self.assertTrue(write_json_to_file_if_different('test2.json', d))
        self.assertFalse(write_json_to_file_if_different('test2.json', d))

    def test_load_json_from_file(self):
        d = {'test': 2}
        write_json_to_file('test.json', d)
        self.assertEqual(read_json_from_file('test.json'), d)

    def test_jsonable_objects(self):
        class MyClass(JSONAble):
            def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
                return []

        self.assertEqual(json_encode(MyClass()), '[]')

    def test_unjsonable_objects(self):
        class MyClass:
            def __init__(self, a):
                self.a = a

        data = json.loads(json_encode(MyClass(5)))
        self.assertEqual(data['a'], 5)

        class MyClass2:
            __slots__ = ('a', 'b')

            def __init__(self, a):
                self.a = a
                self.b = {'test': 'test2'}

        data = json.loads(json_encode(MyClass2(5)))
        self.assertEqual(data['a'], 5)
        self.assertEqual(data['b']['test'], 'test2')



class SetZip:
    """
    An object which zips a bunch of sets together.

    Ie. checks for inclusion by checking each set.

    Also supports len and iteration protocol.

    You can also add extra sets:

    >>> c = SetZip()
    >>> c += set([1, 2, 3])

    Provided arguments must implement contains, length and iter.
    """

    def __iadd__(self, other: set):
        self.args.append(other)
        return self

    def __init__(self, *args: set):
        self.args = list(args)

    def __contains__(self, item) -> bool:
        return any(item in arg for arg in self.args)

    def __len__(self) -> int:
        return sum(len(arg) for arg in self.args)

    def __iter__(self):
        for arg in self.args:
            yield from arg



from satella.os.daemon import daemonize
from satella.os.misc import suicide, is_running_as_root, whereis, safe_listdir
from satella.os.pidlock import PIDFileLock
from satella.os.signals import hang_until_sig

__all__ = [
    'daemonize', 'whereis', 'safe_listdir',
    'PIDFileLock', 'hang_until_sig',
    'suicide', 'is_running_as_root'
]



import collections
import copy
import typing as tp

from satella.coding.typing import T

ITER_KEYS = 0
ITER_VALUES = 1
ITER_ITEMS = 2


class DictDeleter:
    """
    Having problems deleting entries from your dict while iterating on them? No problem. Just swap
    the following:

    >>> keys_to_delete = []
    >>> for key, value in my_dict.items():
    >>>     if value.should_delete():
    >>>         keys_to_delete.append(key)
    >>> for key in keys_to_delete:
    >>>     del my_dict[key]

    With the following:

    >>> with DictDeleter(my_list) as ld:
    >>>     for key, value in ld.items():
    >>>         if value.should_delete():
    >>>             ld.delete()

    Note that a single DictDeleter running from a single context must be iterated on by only a
    single Thread as it keeps the state of iterator in itself, to prevent allocating new objects
    and slowing things down too much.

    This allocates only a single object per a call to delete().
    """
    __slots__ = ('dict_to_process', 'current_iterator', 'keys_to_delete', 'iter_mode',
                 'current_key')

    def __init__(self, dict_to_process: collections.abc.MutableMapping):
        self.dict_to_process = dict_to_process
        self.iter_mode = ITER_KEYS

    def __enter__(self):
        self.keys_to_delete = set()
        return self

    def __iter__(self):
        self.current_iterator = iter(self.dict_to_process.items())
        return self

    def __next__(self):
        key, value = next(self.current_iterator)  # raises StopIteration
        self.current_key = key
        if self.iter_mode == ITER_ITEMS:
            return key, value
        elif self.iter_mode == ITER_VALUES:
            return value
        elif self.iter_mode == ITER_KEYS:
            return key

    def items(self) -> 'DictDeleter':
        self.iter_mode = ITER_ITEMS
        return self

    def keys(self) -> 'DictDeleter':
        self.iter_mode = ITER_KEYS
        return self

    def values(self) -> 'DictDeleter':
        self.iter_mode = ITER_VALUES
        return self

    def delete(self) -> None:
        self.keys_to_delete.add(self.current_key)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for key in self.keys_to_delete:
            del self.dict_to_process[key]
        return False


DIR_FORWARD = 0
DIR_BACKWARD = 1


class ListDeleter(tp.Generic[T]):
    """
    Having problems deleting entries from your list while iterating on them? No problem. Just swap
    the following:

    >>> entries_to_delete = []
    >>> for entry in my_list:
    >>>     if entry.should_delete():
    >>>         entries_to_delete.append(entry)
    >>> for entry in entries_to_delete:
    >>>     my_list.remove(entry)

    With the following:

    >>> with ListDeleter(my_list) as ld:
    >>>     for entry in ld:
    >>>         if entry.should_delete():
    >>>             ld.delete()

    You can also use the alternative syntax of:
    >>> ld = ListDeleter(my_list)
    >>> while True:
    >>>     try:
    >>>         v = ld.next()
    >>>     except StopIteration:
    >>>         break
    >>>     if condition(v):
    >>>         ld.delete()
    >>> ld.remove_items()

    Note that a single ListDeleter running from a single context must be iterated on by only a
    single Thread as it keeps the state of iterator in itself, to prevent allocating new objects
    and slowing things down too much.

    Note that calling reversed() on this will reset the pointer to the end of the list
    or the beginning of the list, respectively.

    This allocates only a single object per a call to delete().

    Calling the list deleter during iteration will yield the element.

    You can pass any type of object here, as long as it supports pop(position) and __getitem__
    """
    __slots__ = 'list_to_process', 'current_index', 'indices_to_delete', 'direction', 'removed'

    def __init__(self, list_to_process: tp.MutableSequence[T], direction: int = DIR_FORWARD):
        self.list_to_process = list_to_process
        # pointer to currently processed element
        self.direction = direction
        self.current_index = -1 if direction == DIR_FORWARD else len(self.list_to_process)
        self.indices_to_delete = set()  # type: tp.Set[int]
        self.removed = False

    def __enter__(self) -> 'ListDeleter':
        return self

    @property
    def value(self) -> T:
        return self()

    def __call__(self) -> T:
        """
        Return current element
        """
        return self.list_to_process[self.current_index]

    def __iter__(self) -> 'ListDeleter':
        self.current_index = -1 if self.direction == DIR_FORWARD else len(self.list_to_process)
        self.indices_to_delete = set()
        return self

    def __next__(self) -> T:
        if self.direction == DIR_BACKWARD and self.current_index == -1:
            raise StopIteration('First element reached')
        if self.direction == DIR_FORWARD and self.current_index == len(self.list_to_process):
            raise StopIteration('Last element reached')
        self.current_index += +1 if self.direction == DIR_FORWARD else -1
        try:
            return self.list_to_process[self.current_index]
        except IndexError:
            raise StopIteration()

    def next(self) -> T:
        """
        :return: the next element
        :raises StopIteration: no more entries
        """
        return next(self)

    def prev(self) -> T:
        """
        Move to previous element, as per ordering.

        :return: the previous element
        :raises StopIteration: list is already at the first element!
        """
        if self.direction == DIR_FORWARD and self.current_index == -1:
            raise StopIteration('Cannot go previous on a first element!')
        if self.direction == DIR_BACKWARD and self.current_index == len(self.list_to_process):
            raise StopIteration('Cannot go previous on a last element!')
        self.current_index += -1 if self.direction == DIR_FORWARD else +1
        try:
            return self.list_to_process[self.current_index]
        except IndexError:
            raise StopIteration()

    def delete(self) -> None:
        self.indices_to_delete.add(self.current_index)

    def __reversed__(self) -> 'ListDeleter':
        """
        This will return a new ListDeleter. Take care that it will break context managers
        such as this:

        >>> with ListDeleter(a):
        >>>     for v in reversed(a):
        >>>         ...

        It is much preferred to do the following in that case:

        >>> with reversed(ListDeleter(a)) as ld:
        >>>     for v in a:
        >>>         if condition(v):
        >>>             ld.delete()
        """
        if self.direction == DIR_BACKWARD:
            return ListDeleter(self.list_to_process)
        else:
            return ListDeleter(self.list_to_process, DIR_BACKWARD)

    def remove_items(self) -> None:
        """
        After all of the items have been marked for deletion, delete them
        """
        if not self.removed:
            index_delta = 0
            for index, element in enumerate(self.list_to_process[:]):
                if index in self.indices_to_delete:
                    self.list_to_process.pop(index - index_delta)
                    index_delta += 1
            self.removed = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove_items()
        return False

    def __copy__(self) -> 'ListDeleter':
        ld = ListDeleter(self.list_to_process)
        ld.direction = self.direction
        ld.current_index = self.current_index
        ld.indices_to_delete = copy.copy(self.indices_to_delete)
        return ld



import copy
import typing as tp

from satella.coding.recast_exceptions import rethrow_as
from satella.configuration.schema import Descriptor, descriptor_from_dict
from satella.exceptions import ConfigurationValidationError


class DictObject(dict):
    """
    A dictionary wrapper that can be accessed by attributes.

    Note that it still remains a completely valid dictionary!

    You can use keys different than strings, but they will be inaccessible as attributes, and
    you will have to do subscription to get them.

    Eg:

    >>> a = DictObject({'test': 5})
    >>> self.assertEqual(a.test, 5)
    >>> self.assertEqual(a['test'], 5)
    """

    def __copy__(self):
        return DictObject(copy.copy(super()))

    def __deepcopy__(self, memo=None):
        return DictObject(copy.deepcopy(super(), memo=memo))

    def __str__(self) -> str:
        return 'DictObject(%s)' % (super().__str__(),)

    def __repr__(self) -> str:
        return 'DictObject(%s)' % (super().__repr__(),)

    @rethrow_as(KeyError, AttributeError)
    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    @rethrow_as(KeyError, AttributeError)
    def __delattr__(self, key):
        del self[key]

    def is_valid_schema(self, schema: tp.Optional[tp.Union[Descriptor, tp.Dict]] = None,
                        **kwarg_schema) -> bool:
        """
        Check if this dictionary conforms to particular schema.

        Schema is either a Descriptor, or a JSON-based schema. See satella.configuration.schema
        for details.
        Schema can be passed as well using kwargs. Note that the schema argument will be ignored
        if kwargs are passed.

        :param schema: schema to verify against
        :return: whether is conformant
        """

        if kwarg_schema:
            schema = kwarg_schema

        if isinstance(schema, Descriptor):
            descriptor = schema
        else:
            descriptor = descriptor_from_dict(schema)

        try:
            descriptor(self)
        except ConfigurationValidationError:
            return False
        else:
            return True


def apply_dict_object(v: tp.Union[tp.Any, tp.Dict]) -> tp.Union[DictObject, tp.Any]:
    """
    Apply DictObject() to every dict inside v.

    This assumes that the only things that will be touched will be nested dicts and lists.

    If you pass a non-dict and a non-list, they will be returned as is.
    """
    if isinstance(v, DictObject):
        return v
    elif isinstance(v, list):
        return [apply_dict_object(x) for x in v]
    elif isinstance(v, dict):
        return DictObject({
            k: apply_dict_object(val) for k, val in v.items()
        })
    else:
        return v



import unittest
from unittest.mock import Mock

from satella.db import transaction


class RealConnection:
    def __init__(self):
        self.cursor_called = 0
        self.commit_called = 0
        self.rollback_called = 0
        self.close_called = 0

    def cursor(self):
        self.cursor_called += 1
        return Mock()

    def rollback(self):
        self.rollback_called += 1

    def close(self):
        self.close_called += 1

    def commit(self):
        self.commit_called += 1


class TestDB(unittest.TestCase):
    def test_db(self):
        conn = RealConnection()
        a = transaction(conn)
        with a as cur:
            self.assertEqual(conn.cursor_called, 1)
            cur.execute('TEST')
        self.assertEqual(conn.commit_called, 1)
        try:
            with a as cur:
                raise ValueError()
        except ValueError:
            self.assertEqual(conn.commit_called, 1)
            self.assertEqual(conn.rollback_called, 1)



import struct
import unittest

from satella.exceptions import NotEnoughBytes
from satella.parsing import BinaryParser


class TestParsing(unittest.TestCase):
    def test_something(self):
        bp = BinaryParser(b'\x00\x00\x01\x02\x00\x00\x00\xFF\x00')
        self.assertEqual(bp.get_bytes(2), b'\x00\x00')
        self.assertEqual(bp.get_structs('BB'), (1, 2))
        self.assertEqual(bp.get_struct(struct.Struct('>L')), 255)
        self.assertRaises(NotEnoughBytes, lambda: bp.get_struct('>L'))
        self.assertRaises(NotEnoughBytes, lambda: bp.get_bytes(5))
        self.assertRaises(NotEnoughBytes, lambda: BinaryParser(b'', 1))
        self.assertEqual(bp.get_remaining_bytes(), b'\x00')
        self.assertEqual(len(bp), 9)
        bp.reset()
        bp2 = bp.get_parser(4)
        self.assertEqual(len(bp2), 4)
        self.assertEqual(bytes(bp2), b'\x00\x00\x01\x02')
        self.assertEqual(bp2.get_struct('>L'), 258)
        self.assertRaises(NotEnoughBytes, lambda: bp2.get_struct('>L'))
        self.assertEqual(bp.pointer, 4)
        bp.reset()
        bp.skip(4)
        bp3 = bp.get_parser(4)
        self.assertEqual(bytes(bp3), b'\x00\x00\x00\xFF')
        self.assertRaises(NotEnoughBytes, lambda: bp2.get_parser(4))
        bp3.skip(3)
        self.assertEqual(bp3.get_remaining_bytes(), b'\xFF')



import unittest

from satella.coding import ListDeleter, DictDeleter


class TestDeleters(unittest.TestCase):

    def test_dict_deleter(self):
        a = {1: 2, 2: 3, 3: 4, 4: 5, 5: 6}
        with DictDeleter(a) as ld:
            for key, value in ld.items():
                if key % 2 == 0:
                    ld.delete()

        self.assertEqual(a, {1: 2, 3: 4, 5: 6})

        a = {1: 2, 2: 3, 3: 4, 4: 5, 5: 6}
        with DictDeleter(a) as ld:
            for key in ld.keys():
                if key % 2 == 0:
                    ld.delete()

        self.assertEqual(a, {1: 2, 3: 4, 5: 6})

        a = {1: 2, 2: 3, 3: 4, 4: 5, 5: 6}
        with DictDeleter(a) as ld:
            for value in ld.values():
                key = value - 1
                if key % 2 == 0:
                    ld.delete()

        self.assertEqual(a, {1: 2, 3: 4, 5: 6})

    def test_list_deleter_reversed(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]
        with reversed(ListDeleter(a)) as ld:
            for v in ld:
                self.assertEqual(v, ld())
                self.assertEqual(v, ld.value)
                if v % 2:
                    ld.delete()
        self.assertEqual(a, [2, 4, 6, 8])

    def test_list_deleter_reversed_prev_next(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]

        with reversed(ListDeleter(a)) as ld:
            # Advance to the end of the list
            self.assertRaises(StopIteration, ld.prev)
            while True:
                # Advance to the start of the list
                try:
                    ld.next()
                except StopIteration:
                    break

            # Advance to the beginning of the list
            while True:
                try:
                    v = ld.prev()
                except StopIteration:
                    break
                if v % 2 == 0:
                    ld.delete()

            self.assertRaises(StopIteration, lambda: ld.prev())
        self.assertEqual(a, [1, 3, 5, 7])

    def test_list_deleter(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]

        with ListDeleter(a) as ld:
            for v in ld:
                if v % 2 == 0:
                    ld.delete()

        self.assertEqual(a, [1, 3, 5, 7])

        a = [1, 2, 3, 4, 5, 6, 7, 8]

        with ListDeleter(a) as ld:
            while True:
                try:
                    v = ld.next()
                except StopIteration:
                    break
                if v % 2 == 0:
                    ld.delete()

        self.assertEqual(a, [1, 3, 5, 7])

        a = [1, 2, 3, 4, 5, 6, 7, 8]

        with ListDeleter(a) as ld:
            # Advance to the end of the list
            self.assertRaises(StopIteration, lambda: ld.prev())
            while True:
                try:
                    v = ld.next()
                    print('upwards', v)
                except StopIteration:
                    break

            while True:
                try:
                    v = ld.prev()
                    print('backwards', v)
                except StopIteration:
                    break
                if v % 2 == 0:
                    ld.delete()

        self.assertEqual(a, [1, 3, 5, 7])
